# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "19/03/2020"

import numpy
from sklearn.decomposition import PCA

from ..core.blindSourceSeparation import IPCA
from .base import Base


class NICA(Base):
    """Compute the non-negative independent components of the linear generative
    model x = A * s.

    Here, x is a p-dimensional observable random vector and s is the latent
    random vector of length num_components, whose components are statistically
    independent and non-negative. The matrix X is assumed to hold n samples
    of x, stacked in rows (shape(X) = (n, p)) or columns (shape(X) = (p, n)),
    which can be specified by the rowvar parameter. In practice, if
    shape(X) = (p, n) (resp. shape(X) = (n, p)) this function solves X = A * S
    (resp. X = S.T * A.T) both for S and A, where A is the so-called mixing
    matrix, with shape (p, num_components), and S is a (num_components, n)
    matrix which contains n samples of the latent source vector, stacked in
    columns.

    This function implements the method presented in:
    `Blind Separation of Positive Sources by Globally Convergent Gradient Search´
    (https://core.ac.uk/download/pdf/76988305.pdf)

    :param num_components: Dimension of s. Number of latent random variables.
    :param float lr: Learning rate of gradient descent.
    :param int max_iter: Maximum number of iterations of gradient descent.
    :param float tol: Tolerance on update at each iteration.
    :param bool rowvar: Whether each row of X corresponds to one of the p
        variables or not.
    :type num_components: Union[uint, None]

    :return: (S, A) if rowvar == True else (S.T, A)

    """

    def __init__(self, data, num_components, chunksize=None, lr=0.03, rowvar=True, indices=None):
        Base.__init__(self, data, num_components=num_components, indices=indices)

        self.lr = lr
        self.rowvar = rowvar

        if chunksize is None:
            model = PCA(num_components, whiten=True)
            if indices is not None:
                self.V = model.fit_transform(self.data[self.indices])
            else:
                self.V = model.fit_transform(self.data)
            self.Z = model.components_
        else:
            self.Z, self.V = IPCA(self.data, chunksize, num_components, whiten=True, indices=indices)

        self.V = self.V.T

        if self.num_components > self.Z.shape[0]:
            self.num_components = self.Z.shape[0]

    def _init_w(self):
        self.W = numpy.eye(self.num_components)

    def _update_w(self):
        Y = numpy.matmul(self.W, self.Z)

        f = numpy.minimum(0, Y)
        f_Y = numpy.matmul(f, Y.transpose())
        E = (f_Y - f_Y.transpose()) / Y.shape[1]
        # Gradient descent
        self.W -= self.lr * numpy.matmul(E, self.W)
        # Symmetric orthogonalization
        M = numpy.matmul(self.W, self.W.transpose())
        vals, vecs = numpy.linalg.eig(M)
        vals, vecs = vals.real, vecs.real

        W_sqrt = vecs / numpy.sqrt(vals)
        W_sqrt = numpy.matmul(W_sqrt, vecs.transpose())
        self.W = numpy.matmul(W_sqrt, self.W)

    def transform(self, max_iter=100, error_step=None):

        Base.transform(self, max_iter=max_iter, error_step=error_step,
                       compute_h=False)

        self.H = numpy.matmul(self.W, self.Z)

        WV = numpy.matmul(self.W, self.V)
        WV_ = numpy.matmul(WV, WV.transpose())
        WV_ = numpy.linalg.inv(WV_)
        self.W = numpy.matmul(WV.transpose(), WV_)
